@@ -30,33 +30,39 @@
 
 from __future__ import division, print_function, absolute_import
 
-import math
+import itertools
 import numpy
+import warnings
+
 from . import _ni_support
 from . import _nd_image
+from ._ni_docstrings import docdict
+from scipy._lib import doccer
 
-import warnings
+# Change the default 'reflect' to 'constant' via modifying a copy of docdict
+docdict_copy = docdict.copy()
+del docdict
+docdict_copy['mode'] = docdict_copy['mode'].replace("Default is 'reflect'",
+                                                    "Default is 'constant'")
+
+docfiller = doccer.filldoc(docdict_copy)
 
 __all__ = ['spline_filter1d', 'spline_filter', 'geometric_transform',
            'map_coordinates', 'affine_transform', 'shift', 'zoom', 'rotate']
 
 
-def _extend_mode_to_code(mode):
-    mode = _ni_support._extend_mode_to_code(mode)
-    return mode
-
-
-def spline_filter1d(input, order=3, axis=-1, output=numpy.float64):
+@docfiller
+def spline_filter1d(input, order=3, axis=-1, output=numpy.float64,
+                    mode='mirror'):
     """
-    Calculates a one-dimensional spline filter along the given axis.
+    Calculate a one-dimensional spline filter along the given axis.
 
     The lines of the array along the given axis are filtered by a
     spline filter. The order of the spline must be >= 2 and <= 5.
 
     Parameters
     ----------
-    input : array_like
-        The input array.
+    %(input)s
     order : int, optional
         The order of the spline, default is 3.
     axis : int, optional
@@ -64,30 +70,42 @@ def spline_filter1d(input, order=3, axis=-1, output=numpy.float64):
         axis.
     output : ndarray or dtype, optional
         The array in which to place the output, or the dtype of the returned
-        array. Default is `numpy.float64`.
+        array. Default is ``numpy.float64``.
+    %(mode)s
 
     Returns
     -------
-    spline_filter1d : ndarray or None
-        The filtered input. If `output` is given as a parameter, None is
-        returned.
+    spline_filter1d : ndarray
+        The filtered input.
 
+    Notes
+    -----
+    All functions in `ndimage.interpolation` do spline interpolation of
+    the input image. If using b-splines of `order > 1`, the input image
+    values have to be converted to b-spline coefficients first, which is
+    done by applying this one-dimensional filter sequentially along all
+    axes of the input. All functions that require b-spline coefficients
+    will automatically filter their inputs, a behavior controllable with
+    the `prefilter` keyword argument. For functions that accept a `mode`
+    parameter, the result will only be correct if it matches the `mode`
+    used when filtering.
     """
     if order < 0 or order > 5:
         raise RuntimeError('spline order not supported')
     input = numpy.asarray(input)
     if numpy.iscomplexobj(input):
         raise TypeError('Complex type not supported')
-    output, return_value = _ni_support._get_output(output, input)
+    output = _ni_support._get_output(output, input)
     if order in [0, 1]:
         output[...] = numpy.array(input)
     else:
+        mode = _ni_support._extend_mode_to_code(mode)
         axis = _ni_support._check_axis(axis, input.ndim)
-        _nd_image.spline_filter1d(input, order, axis, output)
-    return return_value
+        _nd_image.spline_filter1d(input, order, axis, output, mode)
+    return output
 
 
-def spline_filter(input, order=3, output=numpy.float64):
+def spline_filter(input, order=3, output=numpy.float64, mode='mirror'):
     """
     Multi-dimensional spline filter.
 
@@ -111,38 +129,23 @@ def spline_filter(input, order=3, output=numpy.float64):
     input = numpy.asarray(input)
     if numpy.iscomplexobj(input):
         raise TypeError('Complex type not supported')
-    output, return_value = _ni_support._get_output(output, input)
+    output = _ni_support._get_output(output, input)
     if order not in [0, 1] and input.ndim > 0:
         for axis in range(input.ndim):
-            spline_filter1d(input, order, axis, output=output)
+            spline_filter1d(input, order, axis, output=output, mode=mode)
             input = output
     else:
         output[...] = input[...]
-    return return_value
-
-
-def _geometric_transform(input, mapping, coordinates, matrix, offset, output,
-                         order, mode, cval, extra_arguments, extra_keywords):
-    """
-    Wrapper around _nd_image.geometric_transform to work around
-    endianness issues
-    """
-    _nd_image.geometric_transform(
-        input, mapping, coordinates, matrix, offset, output,
-        order, mode, cval, extra_arguments, extra_keywords)
-
-    if output is not None and not output.dtype.isnative:
-        output.byteswap(True)
-
     return output
 
 
+@docfiller
 def geometric_transform(input, mapping, output_shape=None,
                         output=None, order=3,
                         mode='constant', cval=0.0, prefilter=True,
                         extra_arguments=(), extra_keywords={}):
     """
-    Apply an arbritrary geometric transform.
+    Apply an arbitrary geometric transform.
 
     The given mapping function is used to find, for each point in the
     output, the corresponding coordinates in the input. The value of the
@@ -151,32 +154,20 @@ def geometric_transform(input, mapping, output_shape=None,
 
     Parameters
     ----------
-    input : array_like
-        The input array.
-    mapping : callable
+    %(input)s
+    mapping : {callable, scipy.LowLevelCallable}
         A callable object that accepts a tuple of length equal to the output
         array rank, and returns the corresponding input coordinates as a tuple
         of length equal to the input array rank.
     output_shape : tuple of ints, optional
         Shape tuple.
-    output : ndarray or dtype, optional
-        The array in which to place the output, or the dtype of the returned
-        array.
+    %(output)s
     order : int, optional
         The order of the spline interpolation, default is 3.
         The order has to be in the range 0-5.
-    mode : str, optional
-        Points outside the boundaries of the input are filled according
-        to the given mode ('constant', 'nearest', 'reflect' or 'wrap').
-        Default is 'constant'.
-    cval : scalar, optional
-        Value used for points outside the boundaries of the input if
-        ``mode='constant'``. Default is 0.0
-    prefilter : bool, optional
-        The parameter prefilter determines if the input is pre-filtered with
-        `spline_filter` before interpolation (necessary for spline
-        interpolation of order > 1).  If False, it is assumed that the input is
-        already filtered. Default is True.
+    %(mode)s
+    %(cval)s
+    %(prefilter)s
     extra_arguments : tuple, optional
         Extra arguments passed to `mapping`.
     extra_keywords : dict, optional
@@ -184,27 +175,72 @@ def geometric_transform(input, mapping, output_shape=None,
 
     Returns
     -------
-    return_value : ndarray or None
-        The filtered input. If `output` is given as a parameter, None is
-        returned.
+    output : ndarray
+        The filtered input.
 
     See Also
     --------
     map_coordinates, affine_transform, spline_filter1d
 
+
+    Notes
+    -----
+    This function also accepts low-level callback functions with one
+    the following signatures and wrapped in `scipy.LowLevelCallable`:
+
+    .. code:: c
+
+       int mapping(npy_intp *output_coordinates, double *input_coordinates,
+                   int output_rank, int input_rank, void *user_data)
+       int mapping(intptr_t *output_coordinates, double *input_coordinates,
+                   int output_rank, int input_rank, void *user_data)
+
+    The calling function iterates over the elements of the output array,
+    calling the callback function at each element. The coordinates of the
+    current output element are passed through ``output_coordinates``. The
+    callback function must return the coordinates at which the input must
+    be interpolated in ``input_coordinates``. The rank of the input and
+    output arrays are given by ``input_rank`` and ``output_rank``
+    respectively.  ``user_data`` is the data pointer provided
+    to `scipy.LowLevelCallable` as-is.
+
+    The callback function must return an integer error status that is zero
+    if something went wrong and one otherwise. If an error occurs, you should
+    normally set the python error status with an informative message
+    before returning, otherwise a default error message is set by the
+    calling function.
+
+    In addition, some other low-level function pointer specifications
+    are accepted, but these are for backward compatibility only and should
+    not be used in new code.
+
     Examples
     --------
-    >>> from scipy import ndimage
+    >>> import numpy as np
+    >>> from scipy.ndimage import geometric_transform
     >>> a = np.arange(12.).reshape((4, 3))
     >>> def shift_func(output_coords):
     ...     return (output_coords[0] - 0.5, output_coords[1] - 0.5)
     ...
-    >>> ndimage.geometric_transform(a, shift_func)
+    >>> geometric_transform(a, shift_func)
     array([[ 0.   ,  0.   ,  0.   ],
            [ 0.   ,  1.362,  2.738],
            [ 0.   ,  4.812,  6.187],
            [ 0.   ,  8.263,  9.637]])
 
+    >>> b = [1, 2, 3, 4, 5]
+    >>> def shift_func(output_coords):
+    ...     return (output_coords[0] - 3,)
+    ...
+    >>> geometric_transform(b, shift_func, mode='constant')
+    array([0, 0, 0, 1, 2])
+    >>> geometric_transform(b, shift_func, mode='nearest')
+    array([1, 1, 1, 1, 2])
+    >>> geometric_transform(b, shift_func, mode='reflect')
+    array([3, 2, 1, 1, 2])
+    >>> geometric_transform(b, shift_func, mode='wrap')
+    array([2, 3, 4, 1, 2])
+
     """
     if order < 0 or order > 5:
         raise RuntimeError('spline order not supported')
@@ -215,18 +251,19 @@ def geometric_transform(input, mapping, output_shape=None,
         output_shape = input.shape
     if input.ndim < 1 or len(output_shape) < 1:
         raise RuntimeError('input and output rank must be > 0')
-    mode = _extend_mode_to_code(mode)
+    mode = _ni_support._extend_mode_to_code(mode)
     if prefilter and order > 1:
         filtered = spline_filter(input, order, output=numpy.float64)
     else:
         filtered = input
-    output, return_value = _ni_support._get_output(output, input,
-                                                   shape=output_shape)
-    _geometric_transform(filtered, mapping, None, None, None, output,
-                         order, mode, cval, extra_arguments, extra_keywords)
-    return return_value
+    output = _ni_support._get_output(output, input, shape=output_shape)
+    _nd_image.geometric_transform(filtered, mapping, None, None, None, output,
+                                  order, mode, cval, extra_arguments,
+                                  extra_keywords)
+    return output
 
 
+@docfiller
 def map_coordinates(input, coordinates, output=None, order=3,
                     mode='constant', cval=0.0, prefilter=True):
     """
@@ -244,28 +281,16 @@ def map_coordinates(input, coordinates, output=None, order=3,
 
     Parameters
     ----------
-    input : ndarray
-        The input array.
+    %(input)s
     coordinates : array_like
         The coordinates at which `input` is evaluated.
-    output : ndarray or dtype, optional
-        The array in which to place the output, or the dtype of the returned
-        array.
+    %(output)s
     order : int, optional
         The order of the spline interpolation, default is 3.
         The order has to be in the range 0-5.
-    mode : str, optional
-        Points outside the boundaries of the input are filled according
-        to the given mode ('constant', 'nearest', 'reflect' or 'wrap').
-        Default is 'constant'.
-    cval : scalar, optional
-        Value used for points outside the boundaries of the input if
-        ``mode='constant'``. Default is 0.0
-    prefilter : bool, optional
-        The parameter prefilter determines if the input is pre-filtered with
-        `spline_filter` before interpolation (necessary for spline
-        interpolation of order > 1).  If False, it is assumed that the input is
-        already filtered. Default is True.
+    %(mode)s
+    %(cval)s
+    %(prefilter)s
 
     Returns
     -------
@@ -314,24 +339,78 @@ def map_coordinates(input, coordinates, output=None, order=3,
         raise RuntimeError('input and output rank must be > 0')
     if coordinates.shape[0] != input.ndim:
         raise RuntimeError('invalid shape for coordinate array')
-    mode = _extend_mode_to_code(mode)
+    mode = _ni_support._extend_mode_to_code(mode)
     if prefilter and order > 1:
         filtered = spline_filter(input, order, output=numpy.float64)
     else:
         filtered = input
-    output, return_value = _ni_support._get_output(output, input,
-                                                   shape=output_shape)
-    _geometric_transform(filtered, None, coordinates, None, None,
-                         output, order, mode, cval, None, None)
-    return return_value
+    output = _ni_support._get_output(output, input,
+                                     shape=output_shape)
+    _nd_image.geometric_transform(filtered, None, coordinates, None, None,
+                                  output, order, mode, cval, None, None)
+    return output
 
 
+@docfiller
 def affine_transform(input, matrix, offset=0.0, output_shape=None,
                      output=None, order=3,
                      mode='constant', cval=0.0, prefilter=True):
     """
     Apply an affine transformation.
 
+    Given an output image pixel index vector ``o``, the pixel value
+    is determined from the input image at position
+    ``np.dot(matrix, o) + offset``.
+
+    This does 'pull' (or 'backward') resampling, transforming the output space
+    to the input to locate data. Affine transformations are often described in
+    the 'push' (or 'forward') direction, transforming input to output. If you
+    have a matrix for the 'push' transformation, use its inverse
+    (:func:`numpy.linalg.inv`) in this function.
+
+    Parameters
+    ----------
+    %(input)s
+    matrix : ndarray
+        The inverse coordinate transformation matrix, mapping output
+        coordinates to input coordinates. If ``ndim`` is the number of
+        dimensions of ``input``, the given matrix must have one of the
+        following shapes:
+
+            - ``(ndim, ndim)``: the linear transformation matrix for each
+              output coordinate.
+            - ``(ndim,)``: assume that the 2D transformation matrix is
+              diagonal, with the diagonal specified by the given value. A more
+              efficient algorithm is then used that exploits the separability
+              of the problem.
+            - ``(ndim + 1, ndim + 1)``: assume that the transformation is
+              specified using homogeneous coordinates [1]_. In this case, any
+              value passed to ``offset`` is ignored.
+            - ``(ndim, ndim + 1)``: as above, but the bottom row of a
+              homogeneous transformation matrix is always ``[0, 0, ..., 1]``,
+              and may be omitted.
+
+    offset : float or sequence, optional
+        The offset into the array where the transform is applied. If a float,
+        `offset` is the same for each axis. If a sequence, `offset` should
+        contain one value for each axis.
+    output_shape : tuple of ints, optional
+        Shape tuple.
+    %(output)s
+    order : int, optional
+        The order of the spline interpolation, default is 3.
+        The order has to be in the range 0-5.
+    %(mode)s
+    %(cval)s
+    %(prefilter)s
+
+    Returns
+    -------
+    affine_transform : ndarray
+        The transformed input.
+
+    Notes
+    -----
     The given matrix and offset are used to find for each point in the
     output the corresponding coordinates in the input by an affine
     transformation. The value of the input at those coordinates is
@@ -339,65 +418,17 @@ def affine_transform(input, matrix, offset=0.0, output_shape=None,
     outside the boundaries of the input are filled according to the given
     mode.
 
-    If a two-dimensional array is supplied to the matrix parameter,
-    given an output image pixel index vector ``o``, the pixel value
-    is determined from the input image at position ``np.dot(matrix,o) + offset``.
+    .. versionchanged:: 0.18.0
+        Previously, the exact interpretation of the affine transformation
+        depended on whether the matrix was supplied as a one-dimensional or
+        two-dimensional array. If a one-dimensional array was supplied
+        to the matrix parameter, the output pixel value at index ``o``
+        was determined from the input image at position
+        ``matrix * (o + offset)``.
 
-    A diagonal matrix can be specified by supplying a one-dimensional
-    array-like to the matrix parameter, in which case a more efficient
-    algorithm is applied.
-
-    .. warning::
-        Currently, the exact interpretation of the affine transformation
-        depends on whether the matrix is supplied as a one-dimensional or
-        two-dimensional array. This is considered a bug, and starting from
-        scipy 0.18.0, the behaviour for one-dimensional matrix arguments will
-        be modified to match the two-dimensional case.
-
-        Until then, if a one-dimensional array is supplied to the matrix
-        parameter, the output pixel value at index ``o`` is determined from
-        the input image at position ``matrix * (o + offset)``.
-
-    Parameters
+    References
     ----------
-    input : ndarray
-        The input array.
-    matrix : ndarray
-        The matrix must be two-dimensional or can also be given as a
-        one-dimensional sequence or array. In the latter case, it is assumed
-        that the matrix is diagonal. A more efficient algorithms is then
-        applied that exploits the separability of the problem.
-    offset : float or sequence, optional
-        The offset into the array where the transform is applied. If a float,
-        `offset` is the same for each axis. If a sequence, `offset` should
-        contain one value for each axis.
-    output_shape : tuple of ints, optional
-        Shape tuple.
-    output : ndarray or dtype, optional
-        The array in which to place the output, or the dtype of the returned
-        array.
-    order : int, optional
-        The order of the spline interpolation, default is 3.
-        The order has to be in the range 0-5.
-    mode : str, optional
-        Points outside the boundaries of the input are filled according
-        to the given mode ('constant', 'nearest', 'reflect' or 'wrap').
-        Default is 'constant'.
-    cval : scalar, optional
-        Value used for points outside the boundaries of the input if
-        ``mode='constant'``. Default is 0.0
-    prefilter : bool, optional
-        The parameter prefilter determines if the input is pre-filtered with
-        `spline_filter` before interpolation (necessary for spline
-        interpolation of order > 1).  If False, it is assumed that the input is
-        already filtered. Default is True.
-
-    Returns
-    -------
-    affine_transform : ndarray or None
-        The transformed input. If `output` is given as a parameter, None is
-        returned.
-
+    .. [1] https://en.wikipedia.org/wiki/Homogeneous_coordinates
     """
     if order < 0 or order > 5:
         raise RuntimeError('spline order not supported')
@@ -408,16 +439,28 @@ def affine_transform(input, matrix, offset=0.0, output_shape=None,
         output_shape = input.shape
     if input.ndim < 1 or len(output_shape) < 1:
         raise RuntimeError('input and output rank must be > 0')
-    mode = _extend_mode_to_code(mode)
+    mode = _ni_support._extend_mode_to_code(mode)
     if prefilter and order > 1:
         filtered = spline_filter(input, order, output=numpy.float64)
     else:
         filtered = input
-    output, return_value = _ni_support._get_output(output, input,
-                                                   shape=output_shape)
+    output = _ni_support._get_output(output, input,
+                                     shape=output_shape)
     matrix = numpy.asarray(matrix, dtype=numpy.float64)
     if matrix.ndim not in [1, 2] or matrix.shape[0] < 1:
         raise RuntimeError('no proper affine matrix provided')
+    if (matrix.ndim == 2 and matrix.shape[1] == input.ndim + 1 and
+            (matrix.shape[0] in [input.ndim, input.ndim + 1])):
+        if matrix.shape[0] == input.ndim + 1:
+            exptd = [0] * input.ndim + [1]
+            if not numpy.all(matrix[input.ndim] == exptd):
+                msg = ('Expected homogeneous transformation matrix with '
+                       'shape %s for image shape %s, but bottom row was '
+                       'not equal to %s' % (matrix.shape, input.shape, exptd))
+                raise ValueError(msg)
+        # assume input is homogeneous coordinate transformation matrix
+        offset = matrix[:input.ndim, input.ndim]
+        matrix = matrix[:input.ndim, :input.ndim]
     if matrix.shape[0] != input.ndim:
         raise RuntimeError('affine matrix has wrong number of rows')
     if matrix.ndim == 2 and matrix.shape[1] != output.ndim:
@@ -433,17 +476,18 @@ def affine_transform(input, matrix, offset=0.0, output_shape=None,
     if matrix.ndim == 1:
         warnings.warn(
             "The behaviour of affine_transform with a one-dimensional "
-            "array supplied for the matrix parameter will change in "
+            "array supplied for the matrix parameter has changed in "
             "scipy 0.18.0."
         )
-        _nd_image.zoom_shift(filtered, matrix, offset, output, order,
+        _nd_image.zoom_shift(filtered, matrix, offset/matrix, output, order,
                              mode, cval)
     else:
-        _geometric_transform(filtered, None, None, matrix, offset,
-                             output, order, mode, cval, None, None)
-    return return_value
+        _nd_image.geometric_transform(filtered, None, None, matrix, offset,
+                                      output, order, mode, cval, None, None)
+    return output
 
 
+@docfiller
 def shift(input, shift, output=None, order=3, mode='constant', cval=0.0,
           prefilter=True):
     """
@@ -455,35 +499,22 @@ def shift(input, shift, output=None, order=3, mode='constant', cval=0.0,
 
     Parameters
     ----------
-    input : ndarray
-        The input array.
-    shift : float or sequence, optional
+    %(input)s
+    shift : float or sequence
         The shift along the axes. If a float, `shift` is the same for each
         axis. If a sequence, `shift` should contain one value for each axis.
-    output : ndarray or dtype, optional
-        The array in which to place the output, or the dtype of the returned
-        array.
+    %(output)s
     order : int, optional
         The order of the spline interpolation, default is 3.
         The order has to be in the range 0-5.
-    mode : str, optional
-        Points outside the boundaries of the input are filled according
-        to the given mode ('constant', 'nearest', 'reflect' or 'wrap').
-        Default is 'constant'.
-    cval : scalar, optional
-        Value used for points outside the boundaries of the input if
-        ``mode='constant'``. Default is 0.0
-    prefilter : bool, optional
-        The parameter prefilter determines if the input is pre-filtered with
-        `spline_filter` before interpolation (necessary for spline
-        interpolation of order > 1).  If False, it is assumed that the input is
-        already filtered. Default is True.
+    %(mode)s
+    %(cval)s
+    %(prefilter)s
 
     Returns
     -------
-    shift : ndarray or None
-        The shifted input. If `output` is given as a parameter, None is
-        returned.
+    shift : ndarray
+        The shifted input.
 
     """
     if order < 0 or order > 5:
@@ -493,21 +524,22 @@ def shift(input, shift, output=None, order=3, mode='constant', cval=0.0,
         raise TypeError('Complex type not supported')
     if input.ndim < 1:
         raise RuntimeError('input and output rank must be > 0')
-    mode = _extend_mode_to_code(mode)
+    mode = _ni_support._extend_mode_to_code(mode)
     if prefilter and order > 1:
         filtered = spline_filter(input, order, output=numpy.float64)
     else:
         filtered = input
-    output, return_value = _ni_support._get_output(output, input)
+    output = _ni_support._get_output(output, input)
     shift = _ni_support._normalize_sequence(shift, input.ndim)
     shift = [-ii for ii in shift]
     shift = numpy.asarray(shift, dtype=numpy.float64)
     if not shift.flags.contiguous:
         shift = shift.copy()
     _nd_image.zoom_shift(filtered, None, shift, output, order, mode, cval)
-    return return_value
+    return output
 
 
+@docfiller
 def zoom(input, zoom, output=None, order=3, mode='constant', cval=0.0,
          prefilter=True):
     """
@@ -517,36 +549,42 @@ def zoom(input, zoom, output=None, order=3, mode='constant', cval=0.0,
 
     Parameters
     ----------
-    input : ndarray
-        The input array.
-    zoom : float or sequence, optional
+    %(input)s
+    zoom : float or sequence
         The zoom factor along the axes. If a float, `zoom` is the same for each
         axis. If a sequence, `zoom` should contain one value for each axis.
-    output : ndarray or dtype, optional
-        The array in which to place the output, or the dtype of the returned
-        array.
+    %(output)s
     order : int, optional
         The order of the spline interpolation, default is 3.
         The order has to be in the range 0-5.
-    mode : str, optional
-        Points outside the boundaries of the input are filled according
-        to the given mode ('constant', 'nearest', 'reflect' or 'wrap').
-        Default is 'constant'.
-    cval : scalar, optional
-        Value used for points outside the boundaries of the input if
-        ``mode='constant'``. Default is 0.0
-    prefilter : bool, optional
-        The parameter prefilter determines if the input is pre-filtered with
-        `spline_filter` before interpolation (necessary for spline
-        interpolation of order > 1).  If False, it is assumed that the input is
-        already filtered. Default is True.
+    %(mode)s
+    %(cval)s
+    %(prefilter)s
 
     Returns
     -------
-    zoom : ndarray or None
-        The zoomed input. If `output` is given as a parameter, None is
-        returned.
+    zoom : ndarray
+        The zoomed input.
 
+    Examples
+    --------
+    >>> from scipy import ndimage, misc
+    >>> import matplotlib.pyplot as plt
+
+    >>> fig = plt.figure()
+    >>> ax1 = fig.add_subplot(121)  # left side
+    >>> ax2 = fig.add_subplot(122)  # right side
+    >>> ascent = misc.ascent()
+    >>> result = ndimage.zoom(ascent, 3.0)
+    >>> ax1.imshow(ascent)
+    >>> ax2.imshow(result)
+    >>> plt.show()
+
+    >>> print(ascent.shape)
+    (512, 512)
+
+    >>> print(result.shape)
+    (1536, 1536)
     """
     if order < 0 or order > 5:
         raise RuntimeError('spline order not supported')
@@ -555,7 +593,7 @@ def zoom(input, zoom, output=None, order=3, mode='constant', cval=0.0,
         raise TypeError('Complex type not supported')
     if input.ndim < 1:
         raise RuntimeError('input and output rank must be > 0')
-    mode = _extend_mode_to_code(mode)
+    mode = _ni_support._extend_mode_to_code(mode)
     if prefilter and order > 1:
         filtered = spline_filter(input, order, output=numpy.float64)
     else:
@@ -573,34 +611,21 @@ def zoom(input, zoom, output=None, order=3, mode='constant', cval=0.0,
                 "the returned array has changed.", UserWarning)
 
     zoom_div = numpy.array(output_shape, float) - 1
-    zoom = (numpy.array(input.shape) - 1) / zoom_div
-
-    # Zooming to non-finite values is unpredictable, so just choose
+    # Zooming to infinite values is unpredictable, so just choose
     # zoom factor 1 instead
-    zoom[~numpy.isfinite(zoom)] = 1
+    zoom = numpy.divide(numpy.array(input.shape) - 1, zoom_div,
+                        out=numpy.ones_like(input.shape, dtype=numpy.float64),
+                        where=zoom_div != 0)
 
-    output, return_value = _ni_support._get_output(output, input,
-                                                   shape=output_shape)
-    zoom = numpy.asarray(zoom, dtype=numpy.float64)
+    output = _ni_support._get_output(output, input,
+                                     shape=output_shape)
     zoom = numpy.ascontiguousarray(zoom)
     _nd_image.zoom_shift(filtered, zoom, None, output, order, mode, cval)
-    return return_value
+    return output
 
 
-def _minmax(coor, minc, maxc):
-    if coor[0] < minc[0]:
-        minc[0] = coor[0]
-    if coor[0] > maxc[0]:
-        maxc[0] = coor[0]
-    if coor[1] < minc[1]:
-        minc[1] = coor[1]
-    if coor[1] > maxc[1]:
-        maxc[1] = coor[1]
-    return minc, maxc
-
-
-def rotate(input, angle, axes=(1, 0), reshape=True,
-           output=None, order=3,
+@docfiller
+def rotate(input, angle, axes=(1, 0), reshape=True, output=None, order=3,
            mode='constant', cval=0.0, prefilter=True):
     """
     Rotate an array.
@@ -610,8 +635,7 @@ def rotate(input, angle, axes=(1, 0), reshape=True,
 
     Parameters
     ----------
-    input : ndarray
-        The input array.
+    %(input)s
     angle : float
         The rotation angle in degrees.
     axes : tuple of 2 ints, optional
@@ -620,109 +644,111 @@ def rotate(input, angle, axes=(1, 0), reshape=True,
     reshape : bool, optional
         If `reshape` is true, the output shape is adapted so that the input
         array is contained completely in the output. Default is True.
-    output : ndarray or dtype, optional
-        The array in which to place the output, or the dtype of the returned
-        array.
+    %(output)s
     order : int, optional
         The order of the spline interpolation, default is 3.
         The order has to be in the range 0-5.
-    mode : str, optional
-        Points outside the boundaries of the input are filled according
-        to the given mode ('constant', 'nearest', 'reflect' or 'wrap').
-        Default is 'constant'.
-    cval : scalar, optional
-        Value used for points outside the boundaries of the input if
-        ``mode='constant'``. Default is 0.0
-    prefilter : bool, optional
-        The parameter prefilter determines if the input is pre-filtered with
-        `spline_filter` before interpolation (necessary for spline
-        interpolation of order > 1).  If False, it is assumed that the input is
-        already filtered. Default is True.
+    %(mode)s
+    %(cval)s
+    %(prefilter)s
 
     Returns
     -------
-    rotate : ndarray or None
-        The rotated input. If `output` is given as a parameter, None is
-        returned.
+    rotate : ndarray
+        The rotated input.
+
+    Examples
+    --------
+    >>> from scipy import ndimage, misc
+    >>> import matplotlib.pyplot as plt
+    >>> fig = plt.figure(figsize=(10, 3))
+    >>> ax1, ax2, ax3 = fig.subplots(1, 3)
+    >>> img = misc.ascent()
+    >>> img_45 = ndimage.rotate(img, 45, reshape=False)
+    >>> full_img_45 = ndimage.rotate(img, 45, reshape=True)
+    >>> ax1.imshow(img, cmap='gray')
+    >>> ax1.set_axis_off()
+    >>> ax2.imshow(img_45, cmap='gray')
+    >>> ax2.set_axis_off()
+    >>> ax3.imshow(full_img_45, cmap='gray')
+    >>> ax3.set_axis_off()
+    >>> fig.set_tight_layout(True)
+    >>> plt.show()
+    >>> print(img.shape)
+    (512, 512)
+    >>> print(img_45.shape)
+    (512, 512)
+    >>> print(full_img_45.shape)
+    (724, 724)
 
     """
-    input = numpy.asarray(input)
+    input_arr = numpy.asarray(input)
+    ndim = input_arr.ndim
+
+    if ndim < 2:
+        raise ValueError('input array should be at least two-dimensional')
+
     axes = list(axes)
-    rank = input.ndim
+
+    if len(axes) != 2:
+        raise ValueError('axes should contain exactly two values')
+
+    if not all([float(ax).is_integer() for ax in axes]):
+        raise ValueError('axes should contain only integer values')
+
     if axes[0] < 0:
-        axes[0] += rank
+        axes[0] += ndim
     if axes[1] < 0:
-        axes[1] += rank
-    if axes[0] < 0 or axes[1] < 0 or axes[0] > rank or axes[1] > rank:
-        raise RuntimeError('invalid rotation plane specified')
-    if axes[0] > axes[1]:
-        axes = axes[1], axes[0]
-    angle = numpy.pi / 180 * angle
-    m11 = math.cos(angle)
-    m12 = math.sin(angle)
-    m21 = -math.sin(angle)
-    m22 = math.cos(angle)
-    matrix = numpy.array([[m11, m12],
-                             [m21, m22]], dtype=numpy.float64)
-    iy = input.shape[axes[0]]
-    ix = input.shape[axes[1]]
+        axes[1] += ndim
+    if axes[0] < 0 or axes[1] < 0 or axes[0] >= ndim or axes[1] >= ndim:
+        raise ValueError('invalid rotation plane specified')
+
+    axes.sort()
+
+    angle_rad = numpy.deg2rad(angle)
+    c, s = numpy.cos(angle_rad), numpy.sin(angle_rad)
+
+    rot_matrix = numpy.array([[c, s],
+                              [-s, c]])
+
+    img_shape = numpy.asarray(input_arr.shape)
+    in_plane_shape = img_shape[axes]
     if reshape:
-        mtrx = numpy.array([[m11, -m21],
-                               [-m12, m22]], dtype=numpy.float64)
-        minc = [0, 0]
-        maxc = [0, 0]
-        coor = numpy.dot(mtrx, [0, ix])
-        minc, maxc = _minmax(coor, minc, maxc)
-        coor = numpy.dot(mtrx, [iy, 0])
-        minc, maxc = _minmax(coor, minc, maxc)
-        coor = numpy.dot(mtrx, [iy, ix])
-        minc, maxc = _minmax(coor, minc, maxc)
-        oy = int(maxc[0] - minc[0] + 0.5)
-        ox = int(maxc[1] - minc[1] + 0.5)
+        # Compute transformed input bounds
+        iy, ix = in_plane_shape
+        out_bounds = rot_matrix @ [[0, 0, iy, iy],
+                                   [0, ix, 0, ix]]
+        # Compute the shape of the transformed input plane
+        out_plane_shape = (out_bounds.ptp(axis=1) + 0.5).astype(int)
     else:
-        oy = input.shape[axes[0]]
-        ox = input.shape[axes[1]]
-    offset = numpy.zeros((2,), dtype=numpy.float64)
-    offset[0] = float(oy) / 2.0 - 0.5
-    offset[1] = float(ox) / 2.0 - 0.5
-    offset = numpy.dot(matrix, offset)
-    tmp = numpy.zeros((2,), dtype=numpy.float64)
-    tmp[0] = float(iy) / 2.0 - 0.5
-    tmp[1] = float(ix) / 2.0 - 0.5
-    offset = tmp - offset
-    output_shape = list(input.shape)
-    output_shape[axes[0]] = oy
-    output_shape[axes[1]] = ox
+        out_plane_shape = img_shape[axes]
+
+    out_center = rot_matrix @ ((out_plane_shape - 1) / 2)
+    in_center = (in_plane_shape - 1) / 2
+    offset = in_center - out_center
+
+    output_shape = img_shape
+    output_shape[axes] = out_plane_shape
     output_shape = tuple(output_shape)
-    output, return_value = _ni_support._get_output(output, input,
-                                                   shape=output_shape)
-    if input.ndim <= 2:
-        affine_transform(input, matrix, offset, output_shape, output,
+
+    output = _ni_support._get_output(output, input_arr, shape=output_shape)
+
+    if ndim <= 2:
+        affine_transform(input_arr, rot_matrix, offset, output_shape, output,
                          order, mode, cval, prefilter)
     else:
-        coordinates = []
-        size = numpy.product(input.shape,axis=0)
-        size //= input.shape[axes[0]]
-        size //= input.shape[axes[1]]
-        for ii in range(input.ndim):
-            if ii not in axes:
-                coordinates.append(0)
-            else:
-                coordinates.append(slice(None, None, None))
-        iter_axes = list(range(input.ndim))
-        iter_axes.reverse()
-        iter_axes.remove(axes[0])
-        iter_axes.remove(axes[1])
-        os = (output_shape[axes[0]], output_shape[axes[1]])
-        for ii in range(size):
-            ia = input[tuple(coordinates)]
-            oa = output[tuple(coordinates)]
-            affine_transform(ia, matrix, offset, os, oa, order, mode,
-                             cval, prefilter)
-            for jj in iter_axes:
-                if coordinates[jj] < input.shape[jj] - 1:
-                    coordinates[jj] += 1
-                    break
-                else:
-                    coordinates[jj] = 0
-    return return_value
+        # If ndim > 2, the rotation is applied over all the planes
+        # parallel to axes
+        planes_coord = itertools.product(
+            *[[slice(None)] if ax in axes else range(img_shape[ax])
+              for ax in range(ndim)])
+
+        out_plane_shape = tuple(out_plane_shape)
+
+        for coordinates in planes_coord:
+            ia = input_arr[coordinates]
+            oa = output[coordinates]
+            affine_transform(ia, rot_matrix, offset, out_plane_shape,
+                             oa, order, mode, cval, prefilter)
+
+    return output
